"""
sintautils
Created by: Samarthya Lykamanuella
Establishment year: 2025

LICENSE NOTICE:
===============

This file is part of sintautils.

sintautils is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

sintautils is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with sintautils. If not, see <https://www.gnu.org/licenses/>.
"""

from .core import AV



"""
sintautils
Created by: Samarthya Lykamanuella
Establishment year: 2025

LICENSE NOTICE:
===============

This file is part of sintautils.

sintautils is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

sintautils is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with sintautils. If not, see <https://www.gnu.org/licenses/>.
"""
from lxml import html
from requests import Session

from .exceptions import InvalidParameterException, InvalidAuthorIDException, \
    InvalidLoginCredentialException, AuthorIDNotFoundException, MalformedDOMException


class UtilBackEnd(object):
    """ The back-end of sintautils that contain static functions and methods. """

    URL_AUTHOR_BOOK = 'https://sinta.kemdikbud.go.id/authorverification/author/profile/%%%?view=book'
    URL_AUTHOR_GARUDA = 'https://sinta.kemdikbud.go.id/authorverification/author/profile/%%%?view=garuda'
    URL_AUTHOR_GSCHOLAR = 'https://sinta.kemdikbud.go.id/authorverification/author/profile/%%%?view=google'
    URL_AUTHOR_IPR = 'https://sinta.kemdikbud.go.id/authorverification/author/profile/%%%?view=ipr'
    URL_AUTHOR_SCOPUS = 'https://sinta.kemdikbud.go.id/authorverification/author/profile/%%%?view=scopus'
    URL_AUTHOR_RESEARCH = 'https://sinta.kemdikbud.go.id/authorverification/author/profile/%%%?view=research'
    URL_AUTHOR_SERVICE = 'https://sinta.kemdikbud.go.id/authorverification/author/profile/%%%?view=service'
    URL_AUTHOR_WOS = 'https://sinta.kemdikbud.go.id/authorverification/author/profile/%%%?view=wos'

    def __init__(self, requests_session: Session, logger: ()):
        super().__init__()
        self.print = logger
        self.s = requests_session

    @staticmethod
    def _book_heuristics_handler(el):
        """ Scrape with validation a given book row's ISBN, author, publisher, and page information.

        :param el: the XPATH element as DOM object to consider for element detection.
        :return: a dict of the following keys: "isbn", "author", "publisher", and "page".
        """
        ret_dict: dict = {}

        # Scraping each <small> element.
        try:
            s1 = el.xpath('./small[1]/text()')[0]
        except IndexError:
            s1 = ''
        try:
            s2 = el.xpath('./small[2]/text()')[0]
        except IndexError:
            s2 = ''
        try:
            s3 = el.xpath('./small[3]/text()')[0]
        except IndexError:
            s3 = ''
        try:
            s4 = el.xpath('./small[4]/text()')[0]
        except IndexError:
            s4 = ''

        # Carry out the heuristics.
        a = ['ISBN : ', 'Author : ']
        b = ['isbn', 'author']
        for i in range(len(a)):
            lookup_word = a[i]
            key = b[i]
            if s1.__contains__(lookup_word):
                ret_dict[key] = s1.replace(lookup_word, '')
            elif s2.__contains__(lookup_word):
                ret_dict[key] = s2.replace(lookup_word, '')
            elif s3.__contains__(lookup_word):
                ret_dict[key] = s3.replace(lookup_word, '')
            elif s4.__contains__(lookup_word):
                ret_dict[key] = s4.replace(lookup_word, '')
            else:
                ret_dict[key] = ''

        # Now for the special case.
        if s1.__contains__('|'):
            x = s1.split('|')
            ret_dict['publisher'] = x[0].strip()
            ret_dict['page'] = x[1].replace('Page', '').strip()
        elif s2.__contains__('|'):
            x = s2.split('|')
            ret_dict['publisher'] = x[0].strip()
            ret_dict['page'] = x[1].replace('Page', '').strip()
        elif s3.__contains__('|'):
            x = s3.split('|')
            ret_dict['publisher'] = x[0].strip()
            ret_dict['page'] = x[1].replace('Page', '').strip()
        elif s4.__contains__('|'):
            x = s4.split('|')
            ret_dict['publisher'] = x[0].strip()
            ret_dict['page'] = x[1].replace('Page', '').strip()
        else:
            ret_dict['publisher'] = ''
            ret_dict['page'] = ''

        return ret_dict

    @staticmethod
    def _garuda_heuristics_handler(el):
        """ Determines whether a given DOM element contains parameters existing in the "Garuda" data.

        :param el: the XPATH element as DOM object to consider for element detection.
        :return: a dict of the following keys: "publisher", "author", "journal", and "doi".
        """
        ret_dict: dict = {}

        # Scraping each <small> element.
        try:
            s1 = el.xpath('./small[1]/text()')[0]
        except IndexError:
            s1 = ''
        try:
            s2 = el.xpath('./small[2]/text()')[0]
        except IndexError:
            s2 = ''
        try:
            s3 = el.xpath('./small[3]/text()')[0]
        except IndexError:
            s3 = ''
        try:
            s4 = el.xpath('./small[4]/text()')[0]
        except IndexError:
            s4 = ''

        # Carry out the heuristics.
        a = ['DOI: ']
        b = ['doi']
        for i in range(len(a)):
            lookup_word = a[i]
            key = b[i]
            if s1.__contains__(lookup_word):
                ret_dict[key] = s1.replace(lookup_word, '')
            elif s2.__contains__(lookup_word):
                ret_dict[key] = s2.replace(lookup_word, '')
            elif s3.__contains__(lookup_word):
                ret_dict[key] = s3.replace(lookup_word, '')
            elif s4.__contains__(lookup_word):
                ret_dict[key] = s4.replace(lookup_word, '')
            else:
                ret_dict[key] = ''

        # Special cases.
        ret_dict['publisher'] = s1
        ret_dict['author'] = s2
        ret_dict['journal'] = s3

        return ret_dict

    @staticmethod
    def _ipr_heuristics_handler(el):
        """ Determines whether a given DOM element contains "No. Permohonan", "Inventor", or "Pemegang Paten" strings.

        :param el: the XPATH element as DOM object to consider for element detection.
        :return: a dict of the following keys: "application_no", "inventor", and "patent_holder".
        """
        ret_dict: dict = {}

        # Scraping each <small> element.
        try:
            s1 = el.xpath('./small[@class="text-muted"][1]/text()')[0]
        except IndexError:
            s1 = ''
        try:
            s2 = el.xpath('./small[@class="text-muted"][2]/text()')[0]
        except IndexError:
            s2 = ''
        try:
            s3 = el.xpath('./small[@class="text-muted"][3]/text()')[0]
        except IndexError:
            s3 = ''

        # Carry out the heuristics.
        a = ['No. Permohonan : ', 'Inventor : ', 'Pemegang Paten : ']
        b = ['application_no', 'inventor', 'patent_holder']
        for i in range(len(a)):
            lookup_word = a[i]
            key = b[i]
            if s1.__contains__(lookup_word):
                ret_dict[key] = s1.replace(lookup_word, '')
            elif s2.__contains__(lookup_word):
                ret_dict[key] = s2.replace(lookup_word, '')
            elif s3.__contains__(lookup_word):
                ret_dict[key] = s3.replace(lookup_word, '')
            else:
                ret_dict[key] = ''

        return ret_dict

    @staticmethod
    def _research_heuristics_handler(el):
        """ Determines whether a given DOM element contains "Funds approved", "Program Hibah", or "Skema" strings.

        :param el: the XPATH element as DOM object to consider for element detection.
        :return: a dict of the following keys: "funds", "program", and "schema".
        """
        ret_dict: dict = {}

        # Scraping each <small> element.
        try:
            s1 = el.xpath('./small[1]/text()')[0]
        except IndexError:
            s1 = ''
        try:
            s2 = el.xpath('./small[2]/text()')[0]
        except IndexError:
            s2 = ''
        try:
            s3 = el.xpath('./small[3]/text()')[0]
        except IndexError:
            s3 = ''
        try:
            s4 = el.xpath('./small[4]/text()')[0]
        except IndexError:
            s4 = ''

        # Carry out the heuristics.
        a = ['Funds approved : ', 'Program Hibah : ', 'Skema : ']
        b = ['funds', 'program', 'schema']
        for i in range(len(a)):
            lookup_word = a[i]
            key = b[i]
            if s1.__contains__(lookup_word):
                ret_dict[key] = s1.replace(lookup_word, '')
            elif s2.__contains__(lookup_word):
                ret_dict[key] = s2.replace(lookup_word, '')
            elif s3.__contains__(lookup_word):
                ret_dict[key] = s3.replace(lookup_word, '')
            elif s4.__contains__(lookup_word):
                ret_dict[key] = s4.replace(lookup_word, '')
            else:
                ret_dict[key] = ''

        return ret_dict

    @staticmethod
    def _service_heuristics_handler(el):
        """ Determines whether a given DOM element contains "Funds approved", "Program Hibah", or "Skema" strings.

        :param el: the XPATH element as DOM object to consider for element detection.
        :return: a dict of the following keys: "funds", "program", and "schema".
        """
        ret_dict: dict = {}

        # Scraping each <small> element.
        try:
            s1 = el.xpath('./small[1]/text()')[0]
        except IndexError:
            s1 = ''
        try:
            s2 = el.xpath('./small[2]/text()')[0]
        except IndexError:
            s2 = ''
        try:
            s3 = el.xpath('./small[3]/text()')[0]
        except IndexError:
            s3 = ''
        try:
            s4 = el.xpath('./small[4]/text()')[0]
        except IndexError:
            s4 = ''

        # Carry out the heuristics.
        a = ['Funds approved : ', 'Program Hibah : ', 'Skema : ']
        b = ['funds', 'program', 'schema']
        for i in range(len(a)):
            lookup_word = a[i]
            key = b[i]
            if s1.__contains__(lookup_word):
                ret_dict[key] = s1.replace(lookup_word, '')
            elif s2.__contains__(lookup_word):
                ret_dict[key] = s2.replace(lookup_word, '')
            elif s3.__contains__(lookup_word):
                ret_dict[key] = s3.replace(lookup_word, '')
            elif s4.__contains__(lookup_word):
                ret_dict[key] = s4.replace(lookup_word, '')
            else:
                ret_dict[key] = ''

        return ret_dict

    @staticmethod
    def validate_author_id(n):
        """ Return true if the input parameter is a valid ID, false if it contains illegal characters. """

        try:
            int(str(n))
            return True

        except ValueError:
            return False

    def _http_get_with_exception(self, url: str, author_id: str = ''):
        """ Send a GET HTTP request and throw errors in case of
        wrong credential or invalid input parameters, such as author_d. """

        r = self.s.get(url)
        if r.url == url:
            return r

        elif 'authorverification/login' in r.url:
            raise InvalidLoginCredentialException()

        elif 'authorverification/author/all' in r.url:
            raise AuthorIDNotFoundException(author_id)

    # noinspection PyDefaultArgument
    def scrape_book(self, author_id: str, out_format: str = 'json', fields: list = ['*']):
        """ Scrape the book information of one, and only one author in SINTA.
        Returns a Python array/list of dictionaries.

        The only supported out (return) formats are as follows:
        - "json" (includes both dict and list)
        - "csv" (stored as pandas DataFrame object)

        The fields that will be returned are as follows:
        - "*"
        - "title"
        - "isbn"
        - "author"
        - "publisher"
        - "page"
        - "year"
        - "location"
        - "thumbnail"
        - "url"
        """
        l: str = str(author_id).strip()

        # Validating the output format.
        if out_format not in ['csv', 'json']:
            raise InvalidParameterException('"out_format" must be one of "csv" and "json"')

        # Validating the author ID.
        if not self.validate_author_id(l):
            raise InvalidAuthorIDException(l)

        # Try to open the author's specific menu page.
        url = self.URL_AUTHOR_BOOK.replace('%%%', l)
        r = self._http_get_with_exception(url, author_id=author_id)

        self.print('Begin scrapping author ID ' + l + '...', 2)

        # Get the list of pagination.
        c = html.fromstring(r.text)
        try:
            s = c.xpath('//div[@class="row"]/div[@class="col-6 text-right"]/small/i/text()')[0]
            s = s.strip().split('|')[0].replace('Page', '').split('of')
            page_to = int(s[1].strip())
            self.print('Detected number of pages: ' + str(page_to), 2)

        except IndexError:
            # This actually means that the author does not have a record. But still...
            self.print(f'Index error in attempting to read the pagination!', 2)
            page_to = 1

        # Preparing an empty list.
        s1, s2, s3, s4, s5, s6, s7, s8, s9 = [], [], [], [], [], [], [], [], []

        # Preparing the temporary URL.
        new_url: str = str()

        # Begin the scraping.
        i = 0
        while i < page_to:
            i += 1
            self.print(f'Scraping page: {i}...', 2)

            # Opens the URL of this page.
            new_url = url + '&page=' + str(i)
            r = self._http_get_with_exception(new_url, author_id=author_id)
            c = html.fromstring(r.text)

            # The base tag.
            base = '//div[@class="row"]/div[@class="col-12"]/table[@class="table"]//tr'
            c_base = c.xpath(base)

            for a in c_base:
                # Title.
                try:
                    s1.append(a.xpath('.//td[2]/a/text()')[0].strip())
                except IndexError:
                    s1.append('')

                # ISBN, author, publisher, page.
                b = a.xpath('.//td[2]')[0]
                x = self._book_heuristics_handler(b)
                s2.append(x['isbn'].replace('|', '').strip())
                s3.append(x['author'].strip())
                s4.append(x['publisher'].strip())
                s5.append(x['page'].strip())

                # Publish year.
                try:
                    s6.append(a.xpath('.//td[3]//strong[1]/text()')[0].strip())
                except IndexError:
                    s6.append('')

                # Publish location.
                try:
                    s7.append(a.xpath('.//td[3]//strong[2]/text()')[0].strip())
                except IndexError:
                    s7.append('')

                # Thumbnail (book's front cover image).
                try:
                    s8.append(a.xpath('.//td[1]/img/@src')[0].strip())
                except IndexError:
                    s8.append('')

                # Url.
                try:
                    s9.append(a.xpath('.//td[2]/a/@href')[0].strip())
                except IndexError:
                    s9.append('')

        self.print(
            f'({len(s1)}, {len(s2)}, {len(s3)}, {len(s4)}, {len(s5)}, {len(s6)}, {len(s7)}, {len(s8)}, {len(s9)})', 2
        )

        if not len(s1) == len(s2) == len(s3) == len(s4) == len(s5) == len(s6) == len(s7) == len(s8) == len(s9):
            raise MalformedDOMException(new_url)

        # Forge the Python dict.
        t = []
        for j in range(len(s1)):
            # Building the JSON dict.
            u = {}

            if '*' in fields or 'title' in fields:
                u['title'] = s1[j]

            if '*' in fields or 'isbn' in fields:
                u['isbn'] = s2[j]

            if '*' in fields or 'author' in fields:
                u['author'] = s3[j]

            if '*' in fields or 'publisher' in fields:
                u['publisher'] = s4[j]

            if '*' in fields or 'page' in fields:
                u['page'] = s5[j]

            if '*' in fields or 'year' in fields:
                u['year'] = s6[j]

            if '*' in fields or 'location' in fields:
                u['location'] = s7[j]

            if '*' in fields or 'thumbnail' in fields:
                u['thumbnail'] = s8[j]

            if '*' in fields or 'url' in fields:
                u['url'] = s9[j]

            t.append(u)

        # Forge the pandas DataFrame object.
        # Building the CSV dict.
        d = {}
        if '*' in fields or 'title' in fields:
            d['title'] = s1

        if '*' in fields or 'isbn' in fields:
            d['isbn'] = s2

        if '*' in fields or 'author' in fields:
            d['author'] = s3

        if '*' in fields or 'publisher' in fields:
            d['publisher'] = s4

        if '*' in fields or 'page' in fields:
            d['page'] = s5

        if '*' in fields or 'year' in fields:
            d['year'] = s6

        if '*' in fields or 'location' in fields:
            d['location'] = s7

        if '*' in fields or 'thumbnail' in fields:
            d['thumbnail'] = s8

        if '*' in fields or 'url' in fields:
            d['url'] = s9

        if out_format == 'json':
            return t
        elif out_format == 'csv':
            return d

    # noinspection PyDefaultArgument
    def scrape_garuda(self, author_id: str, out_format: str = 'json', fields: list = ['*']):
        """ Scrape the Garuda journal information of one, and only one author in SINTA.
        Returns a Python array/list of dictionaries.

        The only supported out (return) formats are as follows:
        - "json" (includes both dict and list)
        - "csv" (stored as pandas DataFrame object)

        The fields that will be returned are as follows:
        - "*"
        - "title"
        - "publisher"
        - "author"
        - "journal"
        - "doi"
        - "year"
        - "quartile"
        - "url"
        """
        l: str = str(author_id).strip()

        # Validating the output format.
        if out_format not in ['csv', 'json']:
            raise InvalidParameterException('"out_format" must be one of "csv" and "json"')

        # Validating the author ID.
        if not self.validate_author_id(l):
            raise InvalidAuthorIDException(l)

        # Try to open the author's specific menu page.
        url = self.URL_AUTHOR_GARUDA.replace('%%%', l)
        r = self._http_get_with_exception(url, author_id=author_id)

        self.print('Begin scrapping author ID ' + l + '...', 2)

        # Get the list of pagination.
        c = html.fromstring(r.text)
        try:
            s = c.xpath('//div[@class="col-12"]/div[2]/div[2]/small/text()')[0]
            s = s.strip().split('|')[0].replace('Page', '').split('of')
            page_to = int(s[1].strip())
            self.print('Detected number of pages: ' + str(page_to), 2)

        except IndexError:
            # This actually means that the author does not have a record. But still...
            self.print(f'Index error in attempting to read the pagination!', 2)
            page_to = 1

        # Preparing an empty list.
        s1, s2, s3, s4, s5, s6, s7, s8 = [], [], [], [], [], [], [], []

        # Preparing the temporary URL.
        new_url: str = str()

        # Begin the scraping.
        i = 0
        while i < page_to:
            i += 1
            self.print(f'Scraping page: {i}...', 2)

            # Opens the URL of this page.
            new_url = url + '&page=' + str(i)
            r = self._http_get_with_exception(new_url, author_id=author_id)
            c = html.fromstring(r.text)

            # The base tag.
            base = '//div[@class="row"]/div[@class="col-12"]//table[@class="table"]//tr'
            c_base = c.xpath(base)

            for a in c_base:
                # Title.
                try:
                    s1.append(a.xpath('.//td[2]/a/text()')[0].strip())
                except IndexError:
                    s1.append('')

                # Publisher, author, journal name, and DOI.
                b = a.xpath('.//td[2]')[0]
                x = self._garuda_heuristics_handler(b)
                s2.append(x['publisher'].strip())
                s3.append(x['author'].strip())
                s4.append(x['journal'].strip())
                s5.append(x['doi'].strip())

                # Year.
                try:
                    s6.append(a.xpath('.//td[3]//strong/text()')[0].strip())
                except IndexError:
                    s6.append('')

                # Quartile.
                try:
                    s7.append(a.xpath('.//td[1]/div/text()')[0].strip())
                except IndexError:
                    s7.append('')

                # Url.
                try:
                    s8.append(a.xpath('.//td[2]/a/@href')[0].strip())
                except IndexError:
                    s8.append('')

        self.print(f'({len(s1)}, {len(s2)}, {len(s3)}, {len(s4)}, {len(s5)}, {len(s6)}, {len(s7)}, {len(s8)})', 2)

        if not len(s1) == len(s2) == len(s3) == len(s4) == len(s5) == len(s6) == len(s7) == len(s8):
            raise MalformedDOMException(new_url)

        # Forge the Python dict.
        t = []
        for j in range(len(s1)):
            # Building the JSON dict.
            u = {}

            if '*' in fields or 'title' in fields:
                u['title'] = s1[j]

            if '*' in fields or 'publisher' in fields:
                u['publisher'] = s2[j]

            if '*' in fields or 'author' in fields:
                u['author'] = s3[j]

            if '*' in fields or 'journal' in fields:
                u['journal'] = s4[j]

            if '*' in fields or 'doi' in fields:
                u['doi'] = s5[j]

            if '*' in fields or 'year' in fields:
                u['year'] = s6[j]

            if '*' in fields or 'quartile' in fields:
                u['quartile'] = s7[j]

            if '*' in fields or 'url' in fields:
                u['url'] = s8[j]

            t.append(u)

        # Forge the pandas DataFrame object.
        # Building the CSV dict.
        d = {}
        if '*' in fields or 'title' in fields:
            d['title'] = s1

        if '*' in fields or 'publisher' in fields:
            d['publisher'] = s2

        if '*' in fields or 'author' in fields:
            d['author'] = s3

        if '*' in fields or 'journal' in fields:
            d['journal'] = s4

        if '*' in fields or 'doi' in fields:
            d['doi'] = s5

        if '*' in fields or 'year' in fields:
            d['year'] = s6

        if '*' in fields or 'quartile' in fields:
            d['quartile'] = s7

        if '*' in fields or 'url' in fields:
            d['url'] = s8

        if out_format == 'json':
            return t
        elif out_format == 'csv':
            return d

    # noinspection PyDefaultArgument
    def scrape_gscholar(self, author_id: str, out_format: str = 'json', fields: list = ['*']):
        """ Scrape the Google Scholar information of one, and only one author in SINTA.
        Returns a Python array/list of dictionaries.

        The only supported out (return) formats are as follows:
        - "json" (includes both dict and list)
        - "csv" (stored as pandas DataFrame object)

        The fields that will be returned are as follows:
        - "*"
        - "title"
        - "author"
        - "journal"
        - "year"
        - "citations"
        - "url"
        """
        l: str = str(author_id).strip()

        # Validating the output format.
        if out_format not in ['csv', 'json']:
            raise InvalidParameterException('"out_format" must be one of "csv" and "json"')

        # Validating the author ID.
        if not self.validate_author_id(l):
            raise InvalidAuthorIDException(l)

        # Try to open the author's specific menu page.
        url = self.URL_AUTHOR_GSCHOLAR.replace('%%%', l)
        r = self._http_get_with_exception(url, author_id=author_id)

        self.print('Begin scrapping author ID ' + l + '...', 2)

        # Get the list of pagination.
        c = html.fromstring(r.text)
        try:
            s = c.xpath('//div[@class="col-md-12"]/div[2]/div[2]/small/text()')[0]
            s = s.strip().split('|')[0].replace('Page', '').split('of')
            page_to = int(s[1].strip())
            self.print('Detected number of pages: ' + str(page_to), 2)

        except IndexError:
            # This actually means that the author does not have a record. But still...
            self.print(f'Index error in attempting to read the pagination!', 2)
            page_to = 1

        # Preparing an empty list.
        s1, s2, s3, s4, s5, s6 = [], [], [], [], [], []

        # Preparing the temporary URL.
        new_url: str = str()

        # Begin the scraping.
        i = 0
        while i < page_to:
            i += 1
            self.print(f'Scraping page: {i}...', 2)

            # Opens the URL of this page.
            new_url = url + '&page=' + str(i)
            r = self._http_get_with_exception(new_url, author_id=author_id)
            c = html.fromstring(r.text)

            # The base tag.
            base = '//div[@class="table-responsive"]/table[@class="table"]//tr'
            c_base = c.xpath(base)

            for a in c_base:
                # Title.
                try:
                    s1.append(a.xpath('.//a/text()')[0].strip())
                except IndexError:
                    s1.append('')

                # Author.
                try:
                    x: str = a.xpath('.//td[@class="text-lg-nowrap text-nowrap"]//small[1]/text()')[0]
                    if x.__contains__('Author :'):
                        s2.append(x.replace('Author :', '').strip())
                    else:
                        # This is publication info, not author info.
                        # The author info must be missing.
                        s2.append('')
                except IndexError:
                    s2.append('')

                # Journal name.
                try:
                    s3.append(a.xpath('.//td[@class="text-lg-nowrap text-nowrap"]//small[2]/text()')[0].strip())
                except IndexError:
                    s3.append('')

                # Publication year.
                try:
                    s4.append(a.xpath('.//td[2]//strong/text()')[0].strip())
                except IndexError:
                    s4.append('')

                # Citations.
                try:
                    s5.append(a.xpath('.//td[3]//strong/text()')[0].strip())
                except IndexError:
                    s5.append('')

                # URL.
                try:
                    s6.append(a.xpath('.//a/@href')[0].strip())
                except IndexError:
                    s6.append('')

        self.print(f'({len(s1)}, {len(s2)}, {len(s3)}, {len(s4)}, {len(s5)}, {len(s6)})', 2)

        if not len(s1) == len(s2) == len(s3) == len(s4) == len(s5) == len(s6):
            raise MalformedDOMException(new_url)

        # Forge the Python dict.
        t = []
        for j in range(len(s1)):
            # Building the JSON dict.
            u = {}

            if '*' in fields or 'title' in fields:
                u['title'] = s1[j]

            if '*' in fields or 'author' in fields:
                u['author'] = s2[j]

            if '*' in fields or 'journal' in fields:
                u['journal'] = s3[j]

            if '*' in fields or 'year' in fields:
                u['year'] = s4[j]

            if '*' in fields or 'citations' in fields:
                u['citations'] = s5[j]

            if '*' in fields or 'url' in fields:
                u['url'] = s6[j]

            t.append(u)

        # Forge the pandas DataFrame object.
        # Building the CSV dict.
        d = {}
        if '*' in fields or 'title' in fields:
            d['title'] = s1

        if '*' in fields or 'author' in fields:
            d['author'] = s2

        if '*' in fields or 'journal' in fields:
            d['journal'] = s3

        if '*' in fields or 'year' in fields:
            d['year'] = s4

        if '*' in fields or 'citations' in fields:
            d['citations'] = s5

        if '*' in fields or 'url' in fields:
            d['url'] = s6

        if out_format == 'json':
            return t
        elif out_format == 'csv':
            return d

    # noinspection PyDefaultArgument
    def scrape_ipr(self, author_id: str, out_format: str = 'json', fields: list = ['*']):
        """ Scrape the IPR information of one, and only one author in SINTA.
        Returns a Python array/list of dictionaries.

        The only supported out (return) formats are as follows:
        - "json" (includes both dict and list)
        - "csv" (stored as pandas DataFrame object)

        The fields that will be returned are as follows:
        - "*"
        - "title"
        - "application_no"
        - "inventor"
        - "patent_holder"
        - "category"
        - "year"
        - "status"
        """
        l: str = str(author_id).strip()

        # Validating the output format.
        if out_format not in ['csv', 'json']:
            raise InvalidParameterException('"out_format" must be one of "csv" and "json"')

        # Validating the author ID.
        if not self.validate_author_id(l):
            raise InvalidAuthorIDException(l)

        # Try to open the author's specific menu page.
        url = self.URL_AUTHOR_IPR.replace('%%%', l)
        r = self._http_get_with_exception(url, author_id=author_id)

        self.print('Begin scrapping author ID ' + l + '...', 2)

        # Get the list of pagination.
        c = html.fromstring(r.text)
        try:
            s = c.xpath('//div[@class="row"]/div[@class="col-6 text-right"]/small/i/text()')[0]
            s = s.strip().split('|')[0].replace('Page', '').split('of')
            page_to = int(s[1].strip())
            self.print('Detected number of pages: ' + str(page_to), 2)

        except IndexError:
            # This actually means that the author does not have a record. But still...
            self.print(f'Index error in attempting to read the pagination!', 2)
            page_to = 1

        # Preparing an empty list.
        s1, s2, s3, s4, s5, s6, s7 = [], [], [], [], [], [], []

        # Preparing the temporary URL.
        new_url: str = str()

        # Begin the scraping.
        i = 0
        while i < page_to:
            i += 1
            self.print(f'Scraping page: {i}...', 2)

            # Opens the URL of this page.
            new_url = url + '&page=' + str(i)
            r = self._http_get_with_exception(new_url, author_id=author_id)
            c = html.fromstring(r.text)

            # The base tag.
            base = '//div[@class="row"]/div[@class="col-12"]/table[@class="table"]//tr'
            c_base = c.xpath(base)

            for a in c_base:
                # Title.
                try:
                    s1.append(a.xpath('.//td[1]/text()')[0].strip())
                except IndexError:
                    s1.append('')

                # Application number, inventor, and patent holder.
                b = a.xpath('.//td[1]')[0]
                x = self._ipr_heuristics_handler(b)
                s2.append(x['application_no'].strip())
                s3.append(x['inventor'].strip())
                s4.append(x['patent_holder'].strip())

                # Category.
                try:
                    s5.append(a.xpath('.//td[2]//strong[1]/text()')[0].strip())
                except IndexError:
                    s5.append('')

                # Application year.
                try:
                    s6.append(a.xpath('.//td[2]//strong[2]/text()')[0].strip())
                except IndexError:
                    s6.append('')

                # Status.
                try:
                    s7.append(a.xpath('.//td[2]//strong[3]/text()')[0].strip())
                except IndexError:
                    s7.append('')

        self.print(f'({len(s1)}, {len(s2)}, {len(s3)}, {len(s4)}, {len(s5)}, {len(s6)}, {len(s7)})', 2)

        if not len(s1) == len(s2) == len(s3) == len(s4) == len(s5) == len(s6) == len(s7):
            raise MalformedDOMException(new_url)

        # Forge the Python dict.
        t = []
        for j in range(len(s1)):
            # Building the JSON dict.
            u = {}

            if '*' in fields or 'title' in fields:
                u['title'] = s1[j]

            if '*' in fields or 'application_no' in fields:
                u['application_no'] = s2[j]

            if '*' in fields or 'inventor' in fields:
                u['inventor'] = s3[j]

            if '*' in fields or 'patent_holder' in fields:
                u['patent_holder'] = s4[j]

            if '*' in fields or 'category' in fields:
                u['category'] = s5[j]

            if '*' in fields or 'year' in fields:
                u['year'] = s6[j]

            if '*' in fields or 'status' in fields:
                u['status'] = s7[j]

            t.append(u)

        # Forge the pandas DataFrame object.
        # Building the CSV dict.
        d = {}
        if '*' in fields or 'title' in fields:
            d['title'] = s1

        if '*' in fields or 'application_no' in fields:
            d['application_no'] = s2

        if '*' in fields or 'inventor' in fields:
            d['inventor'] = s3

        if '*' in fields or 'patent_holder' in fields:
            d['patent_holder'] = s4

        if '*' in fields or 'category' in fields:
            d['category'] = s5

        if '*' in fields or 'year' in fields:
            d['year'] = s6

        if '*' in fields or 'status' in fields:
            d['status'] = s7

        if out_format == 'json':
            return t
        elif out_format == 'csv':
            return d

    # noinspection PyDefaultArgument
    def scrape_research(self, author_id: str, out_format: str = 'json', fields: list = ['*']):
        """ Scrape the research information of one, and only one author in SINTA.
        Returns a Python array/list of dictionaries.

        The only supported out (return) formats are as follows:
        - "json" (includes both dict and list)
        - "csv" (stored as pandas DataFrame object)

        The fields that will be returned are as follows:
        - "*"
        - "title"
        - "funds"
        - "program"
        - "schema"
        - "year"
        - "membership"
        - "url"
        """
        l: str = str(author_id).strip()

        # Validating the output format.
        if out_format not in ['csv', 'json']:
            raise InvalidParameterException('"out_format" must be one of "csv" and "json"')

        # Validating the author ID.
        if not self.validate_author_id(l):
            raise InvalidAuthorIDException(l)

        # Try to open the author's specific menu page.
        url = self.URL_AUTHOR_RESEARCH.replace('%%%', l)
        r = self._http_get_with_exception(url, author_id=author_id)

        self.print('Begin scrapping author ID ' + l + '...', 2)

        # Get the list of pagination.
        c = html.fromstring(r.text)
        try:
            s = c.xpath('//div[@class="row"]/div[@class="col-6 text-right"]/small/i/text()')[0]
            s = s.strip().split('|')[0].replace('Page', '').split('of')
            page_to = int(s[1].strip())
            self.print('Detected number of pages: ' + str(page_to), 2)

        except IndexError:
            # This actually means that the author does not have a record. But still...
            self.print(f'Index error in attempting to read the pagination!', 2)
            page_to = 1

        # Preparing an empty list.
        s1, s2, s3, s4, s5, s6, s7 = [], [], [], [], [], [], []

        # Preparing the temporary URL.
        new_url: str = str()

        # Begin the scraping.
        i = 0
        while i < page_to:
            i += 1
            self.print(f'Scraping page: {i}...', 2)

            # Opens the URL of this page.
            new_url = url + '&page=' + str(i)
            r = self._http_get_with_exception(new_url, author_id=author_id)
            c = html.fromstring(r.text)

            # The base tag.
            base = '//div[@class="row"]/div[@class="col-12"]/table[@class="table"]//tr'
            c_base = c.xpath(base)

            for a in c_base:
                # Title.
                try:
                    s1.append(a.xpath('.//td[1]/a/text()')[0].strip())
                except IndexError:
                    s1.append('')

                # Funds, program, and schema.
                b = a.xpath('.//td[1]')[0]
                x = self._research_heuristics_handler(b)
                s2.append(x['funds'].strip())
                s3.append(x['program'].strip())
                s4.append(x['schema'].strip())

                # Publish year.
                try:
                    s5.append(a.xpath('.//td[2]//strong/text()')[0].strip())
                except IndexError:
                    s5.append('')

                # Membership role.
                try:
                    s6.append(a.xpath('.//td[3]//strong/text()')[0].strip())
                except IndexError:
                    s6.append('')

                # Url.
                try:
                    s7.append(a.xpath('.//td[1]/a/@href')[0].strip())
                except IndexError:
                    s7.append('')

        self.print(f'({len(s1)}, {len(s2)}, {len(s3)}, {len(s4)}, {len(s5)}, {len(s6)}, {len(s7)})', 2)

        if not len(s1) == len(s2) == len(s3) == len(s4) == len(s5) == len(s6) == len(s7):
            raise MalformedDOMException(new_url)

        # Forge the Python dict.
        t = []
        for j in range(len(s1)):
            # Building the JSON dict.
            u = {}

            if '*' in fields or 'title' in fields:
                u['title'] = s1[j]

            if '*' in fields or 'funds' in fields:
                u['funds'] = s2[j]

            if '*' in fields or 'program' in fields:
                u['program'] = s3[j]

            if '*' in fields or 'schema' in fields:
                u['schema'] = s4[j]

            if '*' in fields or 'year' in fields:
                u['year'] = s5[j]

            if '*' in fields or 'membership' in fields:
                u['membership'] = s6[j]

            if '*' in fields or 'url' in fields:
                u['url'] = s7[j]

            t.append(u)

        # Forge the pandas DataFrame object.
        # Building the CSV dict.
        d = {}
        if '*' in fields or 'title' in fields:
            d['title'] = s1

        if '*' in fields or 'funds' in fields:
            d['funds'] = s2

        if '*' in fields or 'program' in fields:
            d['program'] = s3

        if '*' in fields or 'schema' in fields:
            d['schema'] = s4

        if '*' in fields or 'year' in fields:
            d['year'] = s5

        if '*' in fields or 'membership' in fields:
            d['membership'] = s6

        if '*' in fields or 'url' in fields:
            d['url'] = s7

        if out_format == 'json':
            return t
        elif out_format == 'csv':
            return d

    # noinspection PyDefaultArgument
    def scrape_scopus(self, author_id: str, out_format: str = 'json', fields: list = ['*']):
        """ Scrape the Scopus information of one, and only one author in SINTA.
        Returns a Python array/list of dictionaries.

        The only supported out (return) formats are as follows:
        - "json" (includes both dict and list)
        - "csv" (stored as pandas DataFrame object)

        The fields that will be returned are as follows:
        - "*"
        - "title"
        - "author"
        - "journal"
        - "type"
        - "year"
        - "citations"
        - "quartile"
        - "url"
        """
        l: str = str(author_id).strip()

        # Validating the output format.
        if out_format not in ['csv', 'json']:
            raise InvalidParameterException('"out_format" must be one of "csv" and "json"')

        # Validating the author ID.
        if not self.validate_author_id(l):
            raise InvalidAuthorIDException(l)

        # Try to open the author's specific menu page.
        url = self.URL_AUTHOR_SCOPUS.replace('%%%', l)
        r = self._http_get_with_exception(url, author_id=author_id)

        self.print('Begin scrapping author ID ' + l + '...', 2)

        # Get the list of pagination.
        c = html.fromstring(r.text)
        try:
            s = c.xpath('//div[@class="col-md-12"]/div[2]/div[2]/small/text()')[0]
            s = s.strip().split('|')[0].replace('Page', '').split('of')
            page_to = int(s[1].strip())
            self.print('Detected number of pages: ' + str(page_to), 2)

        except IndexError:
            # This actually means that the author does not have a record. But still...
            self.print(f'Index error in attempting to read the pagination!', 2)
            page_to = 1

        # Preparing an empty list.
        s1, s2, s3, s4, s5, s6, s7, s8 = [], [], [], [], [], [], [], []

        # Preparing the temporary URL.
        new_url: str = str()

        # Begin the scraping.
        i = 0
        while i < page_to:
            i += 1
            self.print(f'Scraping page: {i}...', 2)

            # Opens the URL of this page.
            new_url = url + '&page=' + str(i)
            r = self._http_get_with_exception(new_url, author_id=author_id)
            c = html.fromstring(r.text)

            # The base tag.
            base = '//div[@class="table-responsive"]/table[@class="table"]//tr'
            c_base = c.xpath(base)

            for a in c_base:
                # Title.
                try:
                    s1.append(a.xpath('.//a/text()')[0].strip())
                except IndexError:
                    s1.append('')

                # Author.
                try:
                    x: str = a.xpath('.//td[@class="text-lg-nowrap text-nowrap"]//small[1]/text()')[0]
                    if x.__contains__('Creator :'):
                        s2.append(x.replace('Creator :', '').strip())
                    else:
                        # This is publication info, not author info.
                        # The author info must be missing.
                        s2.append('')
                except IndexError:
                    s2.append('')

                # Journal name.
                try:
                    s3.append(a.xpath('.//td[@class="text-lg-nowrap text-nowrap"]//small[2]/text()')[0].strip())
                except IndexError:
                    s3.append('')

                # Publication type.
                try:
                    s4.append(a.xpath('.//td[3]//strong[1]/text()')[0].strip())
                except IndexError:
                    s4.append('')

                # Publication year.
                try:
                    s5.append(a.xpath('.//td[3]//strong[2]/text()')[0].strip())
                except IndexError:
                    s5.append('')

                # Citations.
                try:
                    s6.append(a.xpath('.//td[4]//strong/text()')[0].strip())
                except IndexError:
                    s6.append('')

                # Quartile.
                try:
                    s7.append(a.xpath('.//td[1]/div/text()')[0].strip())
                except IndexError:
                    s7.append('')

                # URL.
                try:
                    s8.append(a.xpath('.//a/@href')[0].strip())
                except IndexError:
                    s8.append('')

        self.print(f'({len(s1)}, {len(s2)}, {len(s3)}, {len(s4)}, {len(s5)}, {len(s6)}, {len(s7)}, {len(s8)})', 2)

        if not len(s1) == len(s2) == len(s3) == len(s4) == len(s5) == len(s6) == len(s7) == len(s8):
            raise MalformedDOMException(new_url)

        # Forge the Python dict.
        t = []
        for j in range(len(s1)):
            # Building the JSON dict.
            u = {}

            if '*' in fields or 'title' in fields:
                u['title'] = s1[j]

            if '*' in fields or 'author' in fields:
                u['author'] = s2[j]

            if '*' in fields or 'journal' in fields:
                u['journal'] = s3[j]

            if '*' in fields or 'type' in fields:
                u['type'] = s4[j]

            if '*' in fields or 'year' in fields:
                u['year'] = s5[j]

            if '*' in fields or 'citations' in fields:
                u['citations'] = s6[j]

            if '*' in fields or 'quartile' in fields:
                u['quartile'] = s7[j]

            if '*' in fields or 'url' in fields:
                u['url'] = s8[j]

            t.append(u)

        # Forge the pandas DataFrame object.
        # Building the CSV dict.
        d = {}
        if '*' in fields or 'title' in fields:
            d['title'] = s1

        if '*' in fields or 'author' in fields:
            d['author'] = s2

        if '*' in fields or 'journal' in fields:
            d['journal'] = s3

        if '*' in fields or 'type' in fields:
            d['type'] = s4

        if '*' in fields or 'year' in fields:
            d['year'] = s5

        if '*' in fields or 'citations' in fields:
            d['citations'] = s6

        if '*' in fields or 'quartile' in fields:
            d['quartile'] = s7

        if '*' in fields or 'url' in fields:
            d['url'] = s8

        if out_format == 'json':
            return t
        elif out_format == 'csv':
            return d

    # noinspection PyDefaultArgument
    def scrape_service(self, author_id: str, out_format: str = 'json', fields: list = ['*']):
        """ Scrape the community service information of one, and only one author in SINTA.
        Returns a Python array/list of dictionaries.

        The only supported out (return) formats are as follows:
        - "json" (includes both dict and list)
        - "csv" (stored as pandas DataFrame object)

        The fields that will be returned are as follows:
        - "*"
        - "title"
        - "funds"
        - "program"
        - "schema"
        - "year"
        - "membership"
        - "url"
        """
        l: str = str(author_id).strip()

        # Validating the output format.
        if out_format not in ['csv', 'json']:
            raise InvalidParameterException('"out_format" must be one of "csv" and "json"')

        # Validating the author ID.
        if not self.validate_author_id(l):
            raise InvalidAuthorIDException(l)

        # Try to open the author's specific menu page.
        url = self.URL_AUTHOR_SERVICE.replace('%%%', l)
        r = self._http_get_with_exception(url, author_id=author_id)

        self.print('Begin scrapping author ID ' + l + '...', 2)

        # Get the list of pagination.
        c = html.fromstring(r.text)
        try:
            s = c.xpath('//div[@class="row"]/div[@class="col-6 text-right"]/small/i/text()')[0]
            s = s.strip().split('|')[0].replace('Page', '').split('of')
            page_to = int(s[1].strip())
            self.print('Detected number of pages: ' + str(page_to), 2)

        except IndexError:
            # This actually means that the author does not have a record. But still...
            self.print(f'Index error in attempting to read the pagination!', 2)
            page_to = 1

        # Preparing an empty list.
        s1, s2, s3, s4, s5, s6, s7 = [], [], [], [], [], [], []

        # Preparing the temporary URL.
        new_url: str = str()

        # Begin the scraping.
        i = 0
        while i < page_to:
            i += 1
            self.print(f'Scraping page: {i}...', 2)

            # Opens the URL of this page.
            new_url = url + '&page=' + str(i)
            r = self._http_get_with_exception(new_url, author_id=author_id)
            c = html.fromstring(r.text)

            # The base tag.
            base = '//div[@class="row"]/div[@class="col-12"]/table[@class="table"]//tr'
            c_base = c.xpath(base)

            for a in c_base:
                # Title.
                try:
                    s1.append(a.xpath('.//td[1]/a/text()')[0].strip())
                except IndexError:
                    s1.append('')

                # Funds, program, and schema.
                b = a.xpath('.//td[1]')[0]
                x = self._service_heuristics_handler(b)
                s2.append(x['funds'].strip())
                s3.append(x['program'].strip())
                s4.append(x['schema'].strip())

                # Publish year.
                try:
                    s5.append(a.xpath('.//td[2]//strong/text()')[0].strip())
                except IndexError:
                    s5.append('')

                # Membership role.
                try:
                    s6.append(a.xpath('.//td[3]//strong/text()')[0].strip())
                except IndexError:
                    s6.append('')

                # Url.
                try:
                    s7.append(a.xpath('.//td[1]/a/@href')[0].strip())
                except IndexError:
                    s7.append('')

        self.print(f'({len(s1)}, {len(s2)}, {len(s3)}, {len(s4)}, {len(s5)}, {len(s6)}, {len(s7)})', 2)

        if not len(s1) == len(s2) == len(s3) == len(s4) == len(s5) == len(s6) == len(s7):
            raise MalformedDOMException(new_url)

        # Forge the Python dict.
        t = []
        for j in range(len(s1)):
            # Building the JSON dict.
            u = {}

            if '*' in fields or 'title' in fields:
                u['title'] = s1[j]

            if '*' in fields or 'funds' in fields:
                u['funds'] = s2[j]

            if '*' in fields or 'program' in fields:
                u['program'] = s3[j]

            if '*' in fields or 'schema' in fields:
                u['schema'] = s4[j]

            if '*' in fields or 'year' in fields:
                u['year'] = s5[j]

            if '*' in fields or 'membership' in fields:
                u['membership'] = s6[j]

            if '*' in fields or 'url' in fields:
                u['url'] = s7[j]

            t.append(u)

        # Forge the pandas DataFrame object.
        # Building the CSV dict.
        d = {}
        if '*' in fields or 'title' in fields:
            d['title'] = s1

        if '*' in fields or 'funds' in fields:
            d['funds'] = s2

        if '*' in fields or 'program' in fields:
            d['program'] = s3

        if '*' in fields or 'schema' in fields:
            d['schema'] = s4

        if '*' in fields or 'year' in fields:
            d['year'] = s5

        if '*' in fields or 'membership' in fields:
            d['membership'] = s6

        if '*' in fields or 'url' in fields:
            d['url'] = s7

        if out_format == 'json':
            return t
        elif out_format == 'csv':
            return d

    # noinspection PyDefaultArgument
    def scrape_wos(self, author_id: str, out_format: str = 'json', fields: list = ['*']):
        """ Scrape the Web of Science (WOS) information of one, and only one author in SINTA.
        Returns a Python array/list of dictionaries.

        The only supported out (return) formats are as follows:
        - "json" (includes both dict and list)
        - "csv" (stored as pandas DataFrame object)

        The fields that will be returned are as follows:
        - "*"
        - "title"
        - "author"
        - "journal"
        - "year"
        - "citations"
        - "quartile"
        - "url"
        """
        l: str = str(author_id).strip()

        # Validating the output format.
        if out_format not in ['csv', 'json']:
            raise InvalidParameterException('"out_format" must be one of "csv" and "json"')

        # Validating the author ID.
        if not self.validate_author_id(l):
            raise InvalidAuthorIDException(l)

        # Try to open the author's specific menu page.
        url = self.URL_AUTHOR_WOS.replace('%%%', l)
        r = self._http_get_with_exception(url, author_id=author_id)

        self.print('Begin scrapping author ID ' + l + '...', 2)

        # Get the list of pagination.
        c = html.fromstring(r.text)
        try:
            s = c.xpath('//div[@class="col-md-12"]/div[2]/div[2]/small/text()')[0]
            s = s.strip().split('|')[0].replace('Page', '').split('of')
            page_to = int(s[1].strip())
            self.print('Detected number of pages: ' + str(page_to), 2)

        except IndexError:
            # This actually means that the author does not have a record. But still...
            self.print(f'Index error in attempting to read the pagination!', 2)
            page_to = 1

        # Preparing an empty list.
        s1, s2, s3, s4, s5, s6, s7 = [], [], [], [], [], [], []

        # Preparing the temporary URL.
        new_url: str = str()

        # Begin the scraping.
        i = 0
        while i < page_to:
            i += 1
            self.print(f'Scraping page: {i}...', 2)

            # Opens the URL of this page.
            new_url = url + '&page=' + str(i)
            r = self._http_get_with_exception(new_url, author_id=author_id)
            c = html.fromstring(r.text)

            # The base tag.
            base = '//div[@class="table-responsive"]/table[@class="table"]//tr'
            c_base = c.xpath(base)

            for a in c_base:
                # Title.
                try:
                    s1.append(a.xpath('.//a/text()')[0].strip())
                except IndexError:
                    s1.append('')

                # Author.
                try:
                    x: str = a.xpath('.//td[@class="text-lg-nowrap text-nowrap"]//small[1]/text()')[0]
                    if x.__contains__('Authors :'):
                        s2.append(x.replace('Authors :', '').strip())
                    else:
                        # This is publication info, not author info.
                        # The author info must be missing.
                        s2.append('')
                except IndexError:
                    s2.append('')

                # Journal name.
                try:
                    s3.append(a.xpath('.//td[@class="text-lg-nowrap text-nowrap"]//small[2]/text()')[0].strip())
                except IndexError:
                    s3.append('')

                # Publication year.
                try:
                    s4.append(a.xpath('.//td[3]//strong/text()')[0].strip()[-4:])
                except IndexError:
                    s4.append('')

                # Citations.
                try:
                    s5.append(a.xpath('.//td[4]//strong/text()')[0].strip())
                except IndexError:
                    s5.append('')

                # Quartile.
                try:
                    s6.append(a.xpath('.//td[1]/div/text()')[0].strip())
                except IndexError:
                    s6.append('')

                # URL.
                try:
                    s7.append(a.xpath('.//a/@href')[0].strip())
                except IndexError:
                    s7.append('')

        self.print(f'({len(s1)}, {len(s2)}, {len(s3)}, {len(s4)}, {len(s5)}, {len(s6)}, {len(s7)})', 2)

        if not len(s1) == len(s2) == len(s3) == len(s4) == len(s5) == len(s6) == len(s7):
            raise MalformedDOMException(new_url)

        # Forge the Python dict.
        t = []
        for j in range(len(s1)):
            # Building the JSON dict.
            u = {}

            if '*' in fields or 'title' in fields:
                u['title'] = s1[j]

            if '*' in fields or 'author' in fields:
                u['author'] = s2[j]

            if '*' in fields or 'journal' in fields:
                u['journal'] = s3[j]

            if '*' in fields or 'year' in fields:
                u['year'] = s4[j]

            if '*' in fields or 'citations' in fields:
                u['citations'] = s5[j]

            if '*' in fields or 'quartile' in fields:
                u['quartile'] = s6[j]

            if '*' in fields or 'url' in fields:
                u['url'] = s7[j]

            t.append(u)

        # Forge the pandas DataFrame object.
        # Building the CSV dict.
        d = {}
        if '*' in fields or 'title' in fields:
            d['title'] = s1

        if '*' in fields or 'author' in fields:
            d['author'] = s2

        if '*' in fields or 'journal' in fields:
            d['journal'] = s3

        if '*' in fields or 'year' in fields:
            d['year'] = s4

        if '*' in fields or 'citations' in fields:
            d['citations'] = s5

        if '*' in fields or 'quartile' in fields:
            d['quartile'] = s6

        if '*' in fields or 'url' in fields:
            d['url'] = s7

        if out_format == 'json':
            return t
        elif out_format == 'csv':
            return d



"""
sintautils
Created by: Samarthya Lykamanuella
Establishment year: 2025

LICENSE NOTICE:
===============

This file is part of sintautils.

sintautils is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

sintautils is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with sintautils. If not, see <https://www.gnu.org/licenses/>.
"""
import copy
from datetime import datetime as dt
import json
from openpyxl import Workbook
import os
import requests as rq
import time

from .exceptions import EmptyFieldException
from .exceptions import InvalidLoginCredentialException
from .exceptions import InvalidParameterException
from .exceptions import NoLoginCredentialsException
from .exceptions import NonStringParameterException
from .backend import UtilBackEnd


# noinspection SpellCheckingInspection
class SintaScraper(object):
    """ The super-class for all SINTA scrapers in sintautils. Do not invoke directly! """

    # Determine the verbosity of loggings and debug messages.
    # Must be of: 0 (no verbose message), 1 (moderate logging level), and 2 (log all events)
    # If less than 0, it will be interpreted as 0.
    # If greater than 2, it will be interpreted as 2.
    verbosity = 1

    # Determine if the timestamp should be given in the logging message.
    log_timestamp = False

    def __init__(self, username: str = '', password: str = ''):
        self.username = username
        self.password = password

        # Initiating the session.
        self.s = rq.Session()

    def print(self, msg, verbose_level: int):
        """ Debug logging with verbosity control.
        :param msg: the message to be logged to the output terminal.
        :param verbose_level: between 0 and 2, determining on which verbosity level this message will be shown.
        """
        if self.verbosity < 0:
            self.verbosity = 0
        elif self.verbosity > 2:
            self.verbosity = 2

        # Do the message logging.
        if verbose_level <= self.verbosity:
            if self.log_timestamp:
                print('::: [' + str(dt.now()) + '] + ' + str(msg))
            else:
                print(str(msg))


class AV(SintaScraper):
    """ The scraper for the SINTA author verification site (https://sinta.kemdikbud.go.id/authorverification). """

    LOGIN_URL = 'https://sinta.kemdikbud.go.id/authorverification/login/do_login'

    def __init__(self, username: str = '', password: str = '', autologin: bool = False):
        if type(username) is not str or type(password) is not str:
            raise NonStringParameterException()

        elif username.strip() == '' or password.strip() == '':
            raise NoLoginCredentialsException()

        else:
            # We got the credential! Now let's proceed.
            super().__init__(username, password)

            # Initializing the back-end.
            self.backend = UtilBackEnd(self.s, self.print)

            if autologin:
                self.login()

    # noinspection PyDefaultArgument,PyTypeChecker
    def _get_dump(self, author_id: str, fields: list = ['*']):
        """ This function actually gets the dump information requested by the function `dump_author()`.

        :param author_id: the author ID of one, and only one author.
        :param fields: the kind of dump information to be returned.
        :return: a dict that corresponds to the data fields requested.

        The return dict, if specified in the `fields` parameter, will have the following keys:
        - "book"
        - "garuda"
        - "gscholar"
        - "ipr"
        - "research"
        - "scopus"
        - "service"
        - "wos"
        """
        ret_dict = {}

        if type(author_id) is not str:
            raise NonStringParameterException()

        if '*' in fields or 'book' in fields:
            self.print('Obtaining book data...', 0)
            ret_dict['book'] = self.get_book(author_id, out_format='json')

        if '*' in fields or 'garuda' in fields:
            self.print('Obtaining garuda data...', 0)
            ret_dict['garuda'] = self.get_garuda(author_id, out_format='json')

        if '*' in fields or 'gscholar' in fields:
            self.print('Obtaining gscholar data...', 0)
            ret_dict['gscholar'] = self.get_gscholar(author_id, out_format='json')

        if '*' in fields or 'ipr' in fields:
            self.print('Obtaining ipr data...', 0)
            ret_dict['ipr'] = self.get_ipr(author_id, out_format='json')

        if '*' in fields or 'research' in fields:
            self.print('Obtaining research data...', 0)
            ret_dict['research'] = self.get_research(author_id, out_format='json')

        if '*' in fields or 'scopus' in fields:
            self.print('Obtaining scopus data...', 0)
            ret_dict['scopus'] = self.get_scopus(author_id, out_format='json')

        if '*' in fields or 'service' in fields:
            self.print('Obtaining service data...', 0)
            ret_dict['service'] = self.get_service(author_id, out_format='json')

        if '*' in fields or 'wos' in fields:
            self.print('Obtaining wos data...', 0)
            ret_dict['wos'] = self.get_wos(author_id, out_format='json')

        return ret_dict

    # noinspection PyDefaultArgument
    def dump_author(
            self,
            author_id: list = [],
            out_folder: str = os.getcwd(),
            out_prefix: str = 'sintautils_dump_author-',
            out_format: str = 'xlsx',
            fields: list = ['*']
    ):
        """ Performs the scraping of an author's all-information data, and save it as a file.

        :param author_id: the list of author IDs to be scraped.
        :param out_folder: the output folder to which all scraping result files will be saved. Defaults to the current working directory.
        :param out_prefix: the prefix file name into which the scraping result(s) will be saved. The default output filename prefix is "sintautils_dump_author-".

        :param out_format: the format of the output result document.

        Currently, the only supported formats are as follows:
        - "csv"
        - "json"
        - "json-pretty"
        - "xlsx"

        :param fields: the types of scraping data to be scraped.

        Currently, the only supported fields are as follows:
        - "*"
        - "book"
        - "garuda"
        - "gscholar"
        - "ipr"
        - "research"
        - "scopus"
        - "service"
        - "wos"

        You can input more than one field. For instance:
        - ["gscholar", "scopus"]
        - ["wos", "research", "scopus"]

        Use asterisk in order to return all fields:
        - ["*"]
        """

        # Validating the fields.
        if fields.__len__() < 1:
            raise EmptyFieldException('book, garuda, gscholar, ipr, research, scopus, service, wos')

        # Validating the output format.
        if type(out_format) is not str or out_format not in ['csv', 'json', 'xlsx']:
            raise InvalidParameterException('"out_format" must be one of "csv", "json", and "xlsx"')

        # This local function carries out the actual dumping of author data.
        def dump(dump_id):
            a: dict = self._get_dump(dump_id, fields=fields)
            if out_format == 'csv':
                for m in a.keys():
                    with open(str(out_folder) + os.sep + str(out_prefix) + str(dump_id) + '-' + m + '.csv', 'w') as fo:
                        b: list = a[m]

                        # Length validation.
                        if b.__len__() < 1:
                            continue

                        # Write the CSV header.
                        headers: str = str()
                        for n in b[0].keys():
                            headers += f'"{n}",'
                        fo.write(headers[:-1] + '\n')

                        # Write the content.
                        for n in b:
                            n: dict = n
                            rows: str = str()
                            for c in n.keys():
                                rows += f'"{n[c]}",'
                            fo.write(rows[:-1] + '\n')

            elif out_format == 'json' or out_format == 'json-pretty':
                b: dict = {
                    'data': {
                        'author_id': dump_id,
                        'scraping_date': int(time.time()),
                        'scraping_result': copy.deepcopy(a)
                    }
                }

                # Saving the JSON file.
                with open(str(out_folder) + os.sep + str(out_prefix) + str(dump_id) + '.json', 'w') as fo:
                    if out_format == 'json-pretty':
                        json.dump(b, fo, indent=4)
                    else:
                        json.dump(b, fo)

            elif out_format == 'xlsx':
                wb = Workbook()
                for m in sorted(a.keys()):
                    ws = wb.create_sheet(m, -1)

                    # Obtaining the data list and validate the data length.
                    b: list = a[m]
                    if b.__len__() < 1:
                        continue

                    # Write the spreadsheet header.
                    headers: list = list(b[0].keys())
                    for i in range(len(headers)):
                        n = headers[i]
                        ws.cell(row=1, column=(i + 1), value=n)

                        # Write the column's content.
                        for j in range(len(b)):
                            c: dict = b[j]
                            # Offset the row number by two, because the first row is header.
                            ws.cell(row=(j + 2), column=(i + 1), value=c[n])

                # Remove sheets that do not represent data type.
                if wb.sheetnames.__len__() > 0:
                    for d in wb.sheetnames:
                        if d not in a.keys():
                            wb.remove(wb[d])

                # Saving the spreadsheet.
                save_file = str(out_folder) + os.sep + str(out_prefix) + str(dump_id) + '.xlsx'
                wb.save(save_file)

        if type(author_id) is str:
            dump(dump_id=author_id)

        elif type(author_id) is list:
            for l in author_id:
                self.print(f'Dumping author data for author ID: {l}...', 0)
                dump(dump_id=l)

        else:
            raise InvalidParameterException('You can only pass list or string into this function')

    # noinspection PyDefaultArgument
    def get_book(self, author_id: list = [], out_format: str = 'csv', fields: list = ['*']):
        """ Performs the scraping of individual author's book data.

        :param author_id: the list of author IDs to be scraped.
        :param out_format: the format of the output result document.

        Currently, the only supported formats are as follows:
        - "csv"
        - "json"

        You can only specify one output format at a time.

        :param fields: the types of field to be scraped.

        Currently, the only supported fields are as follows:
        - "*"
        - "title"
        - "isbn"
        - "author"
        - "publisher"
        - "page"
        - "year"
        - "location"
        - "thumbnail"
        - "url"

        You can input more than one field. For instance:
        - ["title", "thumbnail"]
        - ["isbn", "year", "thumbnail"]

        Use asterisk in order to return all fields:
        - ["*"]
        """

        if type(author_id) is str:
            a = self.backend.scrape_book(author_id=str(author_id), out_format=out_format, fields=fields)

        elif type(author_id) is list:
            a = {}
            for l in author_id:
                self.print(f'Scraping for author ID: {l}...', 1)
                a[l] = self.backend.scrape_book(author_id=l, out_format=out_format, fields=fields)

        else:
            raise InvalidParameterException('You can only pass list or string into this function')

        return a

    # noinspection PyDefaultArgument
    def get_garuda(self, author_id: list = [], out_format: str = 'csv', fields: list = ['*']):
        """ Performs the scraping of individual author's Garuda data.

        :param author_id: the list of author IDs to be scraped.
        :param out_format: the format of the output result document.

        Currently, the only supported formats are as follows:
        - "csv"
        - "json"

        You can only specify one output format at a time.

        :param fields: the types of field to be scraped.

        Currently, the only supported fields are as follows:
        - "*"
        - "title"
        - "publisher"
        - "author"
        - "journal"
        - "doi"
        - "year"
        - "quartile"
        - "url"

        You can input more than one field. For instance:
        - ["journal", "url"]
        - ["quartile", "doi", "year"]

        Use asterisk in order to return all fields:
        - ["*"]
        """

        if type(author_id) is str:
            a = self.backend.scrape_garuda(author_id=str(author_id), out_format=out_format, fields=fields)

        elif type(author_id) is list:
            a = {}
            for l in author_id:
                self.print(f'Scraping for author ID: {l}...', 1)
                a[l] = self.backend.scrape_garuda(author_id=l, out_format=out_format, fields=fields)

        else:
            raise InvalidParameterException('You can only pass list or string into this function')

        return a

    # noinspection PyDefaultArgument
    def get_gscholar(self, author_id: list = [], out_format: str = 'csv', fields: list = ['*']):
        """ Performs the scraping of individual author's Google Scholar data.

        :param author_id: the list of author IDs to be scraped.
        :param out_format: the format of the output result document.

        Currently, the only supported formats are as follows:
        - "csv"
        - "json"

        You can only specify one output format at a time.

        :param fields: the types of field to be scraped.

        Currently, the only supported fields are as follows:
        - "*"
        - "title"
        - "author"
        - "journal"
        - "year"
        - "citations"
        - "url"

        You can input more than one field. For instance:
        - ["journal", "url"]
        - ["title", "author", "year"]

        Use asterisk in order to return all fields:
        - ["*"]
        """

        if type(author_id) is str:
            a = self.backend.scrape_gscholar(author_id=str(author_id), out_format=out_format, fields=fields)

        elif type(author_id) is list:
            a = {}
            for l in author_id:
                self.print(f'Scraping for author ID: {l}...', 1)
                a[l] = self.backend.scrape_gscholar(author_id=l, out_format=out_format, fields=fields)

        else:
            raise InvalidParameterException('You can only pass list or string into this function')

        return a

    # noinspection PyDefaultArgument
    def get_ipr(self, author_id: list = [], out_format: str = 'csv', fields: list = ['*']):
        """ Performs the scraping of individual author's IPR data.

        :param author_id: the list of author IDs to be scraped.
        :param out_format: the format of the output result document.

        Currently, the only supported formats are as follows:
        - "csv"
        - "json"

        You can only specify one output format at a time.

        :param fields: the types of field to be scraped.

        Currently, the only supported fields are as follows:
        - "*"
        - "title"
        - "application_no"
        - "inventor"
        - "patent_holder"
        - "category"
        - "year"
        - "status"

        You can input more than one field. For instance:
        - ["inventor", "status"]
        - ["title", "patent_holder", "status"]

        Use asterisk in order to return all fields:
        - ["*"]
        """

        if type(author_id) is str:
            a = self.backend.scrape_ipr(author_id=str(author_id), out_format=out_format, fields=fields)

        elif type(author_id) is list:
            a = {}
            for l in author_id:
                self.print(f'Scraping for author ID: {l}...', 1)
                a[l] = self.backend.scrape_ipr(author_id=l, out_format=out_format, fields=fields)

        else:
            raise InvalidParameterException('You can only pass list or string into this function')

        return a

    # noinspection PyDefaultArgument
    def get_research(self, author_id: list = [], out_format: str = 'csv', fields: list = ['*']):
        """ Performs the scraping of individual author's research data.

        :param author_id: the list of author IDs to be scraped.
        :param out_format: the format of the output result document.

        Currently, the only supported formats are as follows:
        - "csv"
        - "json"

        You can only specify one output format at a time.

        :param fields: the types of field to be scraped.

        Currently, the only supported fields are as follows:
        - "*"
        - "title"
        - "funds"
        - "program"
        - "schema"
        - "year"
        - "membership"
        - "url"

        You can input more than one field. For instance:
        - ["program", "membership"]
        - ["title", "funds", "membership"]

        Use asterisk in order to return all fields:
        - ["*"]
        """

        if type(author_id) is str:
            a = self.backend.scrape_research(author_id=str(author_id), out_format=out_format, fields=fields)

        elif type(author_id) is list:
            a = {}
            for l in author_id:
                self.print(f'Scraping for author ID: {l}...', 1)
                a[l] = self.backend.scrape_research(author_id=l, out_format=out_format, fields=fields)

        else:
            raise InvalidParameterException('You can only pass list or string into this function')

        return a

    # noinspection PyDefaultArgument
    def get_scopus(self, author_id: list = [], out_format: str = 'csv', fields: list = ['*']):
        """ Performs the scraping of individual author's scopus data.

        :param author_id: the list of author IDs to be scraped.
        :param out_format: the format of the output result document.

        Currently, the only supported formats are as follows:
        - "csv"
        - "json"

        You can only specify one output format at a time.

        :param fields: the types of field to be scraped.

        Currently, the only supported fields are as follows:
        - "*"
        - "title"
        - "author"
        - "journal"
        - "type"
        - "year"
        - "citations"
        - "quartile"
        - "url"

        You can input more than one field. For instance:
        - ["journal", "url"]
        - ["quartile", "citations", "year"]

        Use asterisk in order to return all fields:
        - ["*"]
        """

        if type(author_id) is str:
            a = self.backend.scrape_scopus(author_id=str(author_id), out_format=out_format, fields=fields)

        elif type(author_id) is list:
            a = {}
            for l in author_id:
                self.print(f'Scraping for author ID: {l}...', 1)
                a[l] = self.backend.scrape_scopus(author_id=l, out_format=out_format, fields=fields)

        else:
            raise InvalidParameterException('You can only pass list or string into this function')

        return a

    # noinspection PyDefaultArgument
    def get_service(self, author_id: list = [], out_format: str = 'csv', fields: list = ['*']):
        """ Performs the scraping of individual author's community service data.

        :param author_id: the list of author IDs to be scraped.
        :param out_format: the format of the output result document.

        Currently, the only supported formats are as follows:
        - "csv"
        - "json"

        You can only specify one output format at a time.

        :param fields: the types of field to be scraped.

        Currently, the only supported fields are as follows:
        - "*"
        - "title"
        - "funds"
        - "program"
        - "schema"
        - "year"
        - "membership"
        - "url"

        You can input more than one field. For instance:
        - ["program", "membership"]
        - ["title", "funds", "membership"]

        Use asterisk in order to return all fields:
        - ["*"]
        """

        if type(author_id) is str:
            a = self.backend.scrape_service(author_id=str(author_id), out_format=out_format, fields=fields)

        elif type(author_id) is list:
            a = {}
            for l in author_id:
                self.print(f'Scraping for author ID: {l}...', 1)
                a[l] = self.backend.scrape_service(author_id=l, out_format=out_format, fields=fields)

        else:
            raise InvalidParameterException('You can only pass list or string into this function')

        return a

    # noinspection PyDefaultArgument
    def get_wos(self, author_id: list = [], out_format: str = 'csv', fields: list = ['*']):
        """ Performs the scraping of individual author's Web of Science (WOS) data.

        :param author_id: the list of author IDs to be scraped.
        :param out_format: the format of the output result document.

        Currently, the only supported formats are as follows:
        - "csv"
        - "json"

        You can only specify one output format at a time.

        :param fields: the types of field to be scraped.

        Currently, the only supported fields are as follows:
        - "*"
        - "title"
        - "author"
        - "journal"
        - "year"
        - "citations"
        - "quartile"
        - "url"

        You can input more than one field. For instance:
        - ["journal", "url"]
        - ["quartile", "citations", "year"]

        Use asterisk in order to return all fields:
        - ["*"]
        """

        if type(author_id) is str:
            a = self.backend.scrape_wos(author_id=str(author_id), out_format=out_format, fields=fields)

        elif type(author_id) is list:
            a = {}
            for l in author_id:
                self.print(f'Scraping for author ID: {l}...', 1)
                a[l] = self.backend.scrape_wos(author_id=l, out_format=out_format, fields=fields)

        else:
            raise InvalidParameterException('You can only pass list or string into this function')

        return a

    def login(self):
        """ Performs the credential login and obtains the session cookie for this account. """
        r = self.s.post(self.LOGIN_URL, data={'username': self.username, 'password': self.password})
        self.print(r.status_code, 2)
        self.print(r.url, 2)

        if 'dashboard' in r.url:
            self.print('Login successful!', 1)
        else:
            raise InvalidLoginCredentialException()



"""
sintautils
Created by: Samarthya Lykamanuella
Establishment year: 2025

LICENSE NOTICE:
===============

This file is part of sintautils.

sintautils is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

sintautils is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with sintautils. If not, see <https://www.gnu.org/licenses/>.
"""


class SintaException(Exception):
    """ The super-class for all exceptions related to sintautils. This exception should not be called directly. """


class AuthorIDNotFoundException(SintaException):
    """ Error raised when the author ID specified is not found in the SINTA database. """
    
    def __init__(self, arg: str = ''):
        self.arg = arg
    
    def __repr__(self):
        return f'Cannot locate author ID {self.arg}.'
    
    __str__ = __repr__


class EmptyFieldException(SintaException):
    """ Error raised when there is no field selection in the scraper function passed. """

    def __init__(self, arg: str = ''):
        self.arg = arg

    def __repr__(self):
        return f'You must specify at least one of the following fields: {self.arg}. Use "*" to select all fields.'

    __str__ = __repr__


class InvalidAuthorIDException(SintaException):
    """ Error raised when the user specifies an invalid (i.e., non-numerical) author ID. """

    def __init__(self, arg: str = ''):
        self.arg = arg

    def __repr__(self):
        return f'Invalid author ID: {self.arg}. You must specify a valid, numerical author ID.'
    
    __str__ = __repr__


class InvalidLoginCredentialException(SintaException):
    """ Error raised when the wrong credentials are passed to the login functions in the scraper. """
    def __repr__(self):
        return 'Either your username or password cannot be used to perform the necessary login.'
    
    __str__ = __repr__


class InvalidParameterException(SintaException):
    """ Error raised when a function's parameter setting is not obeyed. """
    
    def __init__(self, arg: str = ''):
        self.arg = arg
    
    def __repr__(self):
        return f'Invaled parameter passed: {self.arg}.'
    
    __str__ = __repr__


class NonStringParameterException(SintaException):
    """ Error raised when non-string parameters are passed to the login function. """
    def __repr__(self):
        return 'You can only pass variables of type str to the method.'
    
    __str__ = __repr__


class NoLoginCredentialsException(SintaException):
    """ Error raised when a `sintautils.scraper.AV` object is created without providing the necessary credential
    information."""
    def __repr__(self):
        return 'You must provide username and password in order to use the AV scraper.'

    __str__ = __repr__


class MalformedDOMException(SintaException):
    """ DOM-related error raised when, e.g., there is an item element
    that does not have a particular property every other item has."""
    
    def __init__(self, arg: str = ''):
        self.arg = arg
    
    def __repr__(self):
        return f'Malformed DOM on URL: {self.arg}.'
    
    __str__ = __repr__
